#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define A 0
#define C 1
#define G 2
#define T 3

double Transitions[2][2];
double Emissions[2][4];

void InitializeA(void);
void InitializeE(void);
void PrintA(void);
void PrintE(void);
void Baum_Welch(char* file);
double Forward(int state, int index, int seq[1000]);
double Backward(int state, int index, int length, int seq[1000]);
int SetSequence(char *sequence, int* seq);
double SequenceProb(int seq[1000],int length);
double ExpectedA(int k, int l, int seq[1000], int length);
double ExpectedE(int k, int b, int seq[1000], int length);
double Max(double a, double b, int *which);
double Viterbi(int length, int seq[1000], int *path);


/* Function Main */

int main(int argc, char *argv[]) {
 
  FILE *fd;
  char* sequence;
  int seq[1000], path[1000], length, i;
  double prob;
  sequence = (char *) malloc(sizeof(char)*1000);
 
  InitializeA();
  InitializeE();
  printf("Initial parameters:\n");
  PrintA();
  PrintE();
  
  if (argc>1) {
    Baum_Welch(argv[1]); 
  } else {
    fprintf(stderr, "Usage: cpg <training_file_name> [<sequence_file_name>]\n");
    exit(-1);
  }

  printf("\n\nParameter after training stage:\n");
  PrintA();
  PrintE();

  if (argc>2) {
    if ((fd = fopen(argv[2],"r")) < 0) {
      fprintf(stderr, "ERROR: Opening sequence file %s.\n",argv[2]);
      exit(-1);
    }
    if (fgets(sequence, 1000, fd) != NULL) {
      printf("\nSequence:\t%s",sequence);
      if ((length = SetSequence(sequence,seq))<0) {
	fprintf(stderr, "ERROR: Incorrect input format in %s.\n",argv[2]);
	exit(-1);
      }
      Viterbi(length, seq, path);
      printf("Opt. path:\t");
      i = 1;
      while (path[i]!=-1) {
	printf("%d",path[i]);
	i++;
      }
      printf("\n");
    }
  }

  exit(0);
}



void InitializeA(void) {
  
  Transitions[0][0] = 0.6; 
  Transitions[0][1] = 1.0 - Transitions[0][0];

  Transitions[1][0] = 0.5; 
  Transitions[1][1] = 1.0 - Transitions[1][0];
}

void InitializeE(void) {
  
  Emissions[0][0] = 0.30;
  Emissions[0][1] = 0.20;
  Emissions[0][2] = 0.20;
  Emissions[0][3] = 0.30;
  Emissions[1][0] = 0.20;
  Emissions[1][1] = 0.30;
  Emissions[1][2] = 0.30;
  Emissions[1][3] = 0.20;


}


void PrintA(void) {
  printf("\nMatrix A:\n");
  printf("\t\tState 1\t\tState 2\n");        
  printf("State 1:\t%f\t%f\n",Transitions[0][0],Transitions[0][1]);
  printf("State 2:\t%f\t%f\n",Transitions[1][0],Transitions[1][1]);
}
void PrintE(void) {
  printf("\nMatrix E:\n");
  printf("\t\tA\t\tC\t\tG\t\tT\n");           
  printf("State 1:\t%f\t%f\t%f\t%f\n",Emissions[0][A],Emissions[0][1],Emissions[0][2],Emissions[0][3]);
  printf("State 2:\t%f\t%f\t%f\t%f\n",Emissions[1][0],Emissions[1][1],Emissions[1][2],Emissions[1][3]);
}


void Baum_Welch(char* file) {
  double a[2][2]; 
  double e[2][4];
  double seqProb;
  int i,j,leng;
  FILE *fd;
  char* sequence;
  int seq[1000];

  sequence = (char *) malloc(sizeof(char)*1000);

  /*Open the training file*/
  if ((fd = fopen(file,"r")) < 0) {
    fprintf(stderr, "ERROR: Opening file %s.\n",file);
    exit(-1);
  }

  /*Initialization*/
  for (i=0; i<2; i++) {
    for (j=0; j<4; j++) {
      if (j<2) {
	a[i][j] = 0.0;
      }
      e[i][j] = 0.0;
    }
  }

  /*Recursion*/
  while (fgets(sequence, 1000, fd) != NULL) {
    if ((leng = SetSequence(sequence,seq))<0) {
      fprintf(stderr, "ERROR: Incorrect input format in %s.\n",file);
      exit(-1);
    }
    seqProb = SequenceProb(seq, leng);
    for (i=0; i<2; i++) {
      for (j=0; j<4; j++) {
	if (j<2) {
	  a[i][j] += (1/seqProb) * ExpectedA(i,j,seq,leng);
	}
	e[i][j] += (1/seqProb) * ExpectedE(i,j,seq,leng);
      }
    } 
  }

  for (i=0; i<2; i++) {
    for (j=0; j<4; j++) {
      if (j<2) {
	Transitions[i][j] = a[i][j]/(a[i][0]+a[i][1]);
      }
      Emissions[i][j] = e[i][j]/(e[i][0]+e[i][1]+e[i][2]+e[i][3]);
    }
  }

  fclose(fd);
}

double ExpectedA(int k, int l, int seq[1000], int length) {
  int i;
  double res = 0.0, f, b;

  for (i=1; i<length; i++) {
    f = Forward(k, i, seq);
    b = Backward(l, i+1, length, seq);
    
    res += (f * b * Transitions[k][l] * Emissions[l][seq[i]]);
  }
  return res;
}

double ExpectedE(int k, int b, int seq[1000], int length) {
  int i;
  double res = 0.0, f, back;

  for (i=1; i<=length; i++) {
    if (seq[i-1]==b) {
      f = Forward(k, i, seq);
      back = Backward(k, i, length, seq);
      res += f * back;
    }
  }
  return res;
}

double Forward(int state, int index, int seq[1000]) {
  double f[2], temp;
  int i;
  
  /*Initialization*/

  f[0] = 1;
  f[1] = 0;

  /*Recursion*/
  
  i = 0;
  while (i <= (index-1)) {
    temp = Emissions[0][seq[i]]*((f[0]*Transitions[0][0])+
			 (f[1]*Transitions[1][0]));  
    f[1] = Emissions[1][seq[i]]*((f[0]*Transitions[0][1])+
		         (f[1]*Transitions[1][1])); 
    f[0] = temp;
    i++;
  }
  return f[state];
}

double SequenceProb(int seq[1000],int length) {
  double p;
  p = Forward(0,length,seq) + Forward(1,length,seq);
  return p;
}

double Backward(int state, int index, int length, int seq[1000]) {
  double b[2], temp;
  int i;

  /*Initialization*/
  b[0] = 1; 
  b[1] = 1;

  /*Recursion*/
  i = length-1;
  while (i >= (index-1)) {
    temp = (Transitions[0][0]*Emissions[0][seq[i]]*b[0])+
           (Transitions[0][1]*Emissions[1][seq[i]]*b[1]);
    b[1] = (Transitions[1][0]*Emissions[0][seq[i]]*b[0])+
           (Transitions[1][1]*Emissions[1][seq[i]]*b[1]);
    b[0] = temp;
    i--;
  }
  return b[state];
}

int SetSequence(char *sequence, int* seq) {
  int i=0, error=0, j, leng=0;

  for (j=0; j<1000; j++) seq[j]=-1;

  while ((sequence[i] != '\n')&&(!error)) {
    if (sequence[i] == 'A') {
      seq[i] = 0;
    }
    else if (sequence[i] == 'C') {
      seq[i] = 1;
    }
    else if (sequence[i] == 'G') {
      seq[i] = 2;
    }
    else if (sequence[i] == 'T') {
      seq[i] = 3;
    }
    else if ((sequence[i] == 'N')||(sequence[i] == ' ')) {
      leng--;
    }
    else {
      error = 1;
    }
    leng++;
    i++;
  }
  if (error) leng = -1;
  return leng;

}



double Viterbi(int length, int seq[1000], int *path) {
  double v[2], temp, res;
  int i,state;
  int ptr0[1000];
  int ptr1[1000];

  /*Initialization*/

  v[0] = log(0.999);
  v[1] = log(0.001);
  ptr0[0] = 0;
  ptr1[0] = 0;

  /*Recursion*/
  
  i = 0;
  while (i < length) {

    temp = log(Emissions[0][seq[i]])+(Max((v[0]+log(Transitions[0][0])),
				     (v[1]+log(Transitions[1][0])),&state));
    
    ptr0[i+1] = state;

    v[1] = log(Emissions[1][seq[i]])+(Max((v[0]+log(Transitions[0][1])),
				     (v[1]+log(Transitions[1][1])),&state)); 
    v[0] = temp;
    ptr1[i+1] = state;
    
    /*printf("v[0] = %f, exp = %f\n", v[0], exp(v[0]));*/

    i++;
  }

  res = Max(v[0],v[1],&path[length]);
  for (i=length-1; i>=0; i--) {
    if (path[i+1]==0) {
      path[i] = ptr0[i];
    } else {
      path[i] = ptr1[i];
    }
  }

  path[length+1] = -1;

  return exp(res);
}

double Max(double a, double b, int *which) {

  double res;

  if (a>=b) {
    *which = 0;
    res = a;
  } else {
    *which = 1;
    res = b;
  }

  return res;
}




